var Obv = require('obv')

var STATES = {
  queried: 1,
  checking: 2,
  checked: 3,
}

function each (obj, fn) {
  for(var k in obj)
    fn(obj[k], k, obj)
}

function isEmpty (o) {
  for(var k in o) return false
  return true
}

function toLength (value) {
  return Array.isArray(value) ? value.length : -1
}

module.exports = function (opts) {
  //opts has {check, process, increment, isQuery, isResponse}
  var state = {}
  var localCbs = {}

  var initialWeight = opts.initialWeight || -1
  var increment  = opts.increment  || function (n) { return Number(n) - 1 }
  var isRequest  = opts.isRequest  || function (value) { return typeof value === 'number' && value < 0 }
  var isResponse = opts.isResponse || function (value) { return !isRequest(value) }
  var isQuery    = opts.isQuery    || function () { return true }
  var compare    = opts.compare    || function (a, b) { return b - a }
  var isUpdate   = opts.isUpdate   || function (_value) { return _value == null }
  var update     = opts.update     || function (_value, value) { return value }
  var maximum = opts.maximum || -3

  if(opts.process) throw new Error('opts.process support removed')

  var obv = Obv()
  obv.set(state)
  function next (fn) {
    if(!fn) obv.set(state)
    else obv.once(fn, false)
  }


  function callback (k, value) {
    if (localCbs[k]) {
      var cbs = localCbs[k]
      delete localCbs[k]
      while (cbs.length) cbs.shift()(null, value)
    }
  }

  obv(function () {
    //XXX: I think there could be some bugs in this because
    //obv.set is triggered from inside the loop (if check/process)
    //callback sync, which can happen.

    each(state, function (item, k) {
      //check the local store when new queries are added
      if(item.state === STATES.queried) {
        item.state = STATES.checking
        opts.check(k, function (err, value) {
          if(err) console.trace(err) // TODO: delete or reject query?
          if(value && !item.value) {
            item.state = STATES.checked
            //UPDATE VALUE
            callback(k, item.value = value)
          }
          else
            item.state = STATES.checked

          obv.set(state)
        })
      }

      if(item.value != null && item.state === STATES.checked) {
        callback(k, item.value) //update(item.value, value))
      }

    })
  })

  function initial (weight) {
    return {
      state: STATES.queried,
      weight: weight,
      value: null,
      requestedBy: {},
      requestedFrom: {},
      respondedTo: {}
    }
  }

  return {
    state: state,
    createStream: function (peerId) {
      return {
        source: function (end, cb) {
          if(end) {
            for(var k in state) {
              // TODO: use hashlru so we don't have to use delete
              delete state[k].respondedTo[peerId]
              delete state[k].requestedFrom[peerId]
            }
            return
          }
          //read the next pieces of data from the state object.
          ;(function read () {
            var data = {}
            for(var k in state) {
              if(state[k].state === STATES.checked) {
                if(
                  //this item has been received & processed, respond to any other peers
                  //idea: make respondedTo be a counter, for queries with multiple answers.
                  state[k].value != null &&
                  state[k].requestedBy[peerId] &&
                  //responded to != state[k].value.length
                  state[k].respondedTo[peerId] != toLength(state[k].value)
                ) {
                  // change bool to integer, and send data since that index
                  if(state[k].respondedTo[peerId] > 0)
                    data[k] = state[k].value.slice(state[k].respondedTo[peerId])
                  else
                    data[k] = state[k].value
                  state[k].respondedTo[peerId] = toLength(state[k].value)
                }
                else if(
                  !state[k].requestedFrom[peerId]
                ) {
                  state[k].requestedFrom[peerId] = true
                  data[k] = state[k].weight //the number of hops, etc
                }
              }
            }
            //next(read) calls read again when something changes in the state.
            if(isEmpty(data)) next(read)
            else cb(null, data)
          })()
        },

        sink: function (read) {
          read(null, function more (end, data) {
            if(end) {
              for(var k in state) {
                delete state[k].requestedBy[peerId]
              }
              //Q: how does the source decide to end?
              //A: the network connection aborts the stream.
              return
            }
            //process this message and possibly update the state.
            var changed = false
            for(var k in data) {
              var value = data[k]
               //ignore invalid requests
              if(!isQuery(k) || !(isRequest(value) || isResponse(value)))
                ;
              else if(isRequest(data[k])) {
                //if we already have seen this query:
                if(compare(data[k], maximum) < 0) {
                  if(!state[k]) {
                    changed = true
                    state[k] = initial(increment(data[k]))
                    state[k].requestedBy[peerId] = true
                  }
                  else if (compare(data[k], state[k].weight) < 0){
                    changed = true
                    state[k].requestedBy[peerId] = true
                    //update the weight if this peer is closer to us
                    state[k].weight = increment(data[k])
                  }
                }
              }
              else if(isResponse(data[k])) {
                if(state[k].state == STATES.checked && isUpdate(state[k].value, data[k], k)) {
                  //what states can it be in here?
                  //what if we are currently processing something and a new response arrives?
                  state[k].value = update(state[k].value, data[k], k)
                  changed = true
                  callback(k, state[k].value)
                }
              }
            }
            if(update) next()
            read(null, more)
          })
        }
      }
    },

    query: function (k, cb) {
      var changed
      //add to state object and update
      if(state[k]) {
        if(state[k].state == STATES.checked) cb(null, state[k].value)
        else (localCbs[k] = localCbs[k] || []).push(cb)
      }
      else {
        changed = true
        state[k] = initial(initialWeight)
        localCbs[k] = [cb]
      }
      if(changed) next()
    },

    progress: function () {
      var prog = {start: 0, current: 0, target: 0}

      for(var k in state) {
        prog.current += state[k].state + (state[k].value != null ? 1 : 0)
        prog.target += STATES.checked + 1
      }
      return prog
    }
  }
}


