var GossipQuery = require('../')
var pull = require('pull-stream')

var tape = require('tape')

var check_cb, process_cb

function once (fn) {
  var called = false
  return function (a,b,c) {
    if(called) throw new Error('called twice')
    called = true
    return fn(a, b, c)
  }
}

function each(obj, iter) {
  if(Array.isArray(obj)) return obj.forEach(iter)
  else return iter(obj)
}

var isArray = Array.isArray

function GQ (store) {
  return GossipQuery({
    check: function (key, cb) {
      check_cb = once(cb)
    },
    isUpdate: function (_value, value, key) {
      if(!_value) return true
      if(isArray(value))
        return value.some(function (e) {
          return !~_value.indexOf(e)
        })
      return !_value || ~_value.indexOf(value)
    },
    update: function (_value, value, key) {
      _value = _value || []
      each(value, function (e) {
        if(!~_value.indexOf(e)) _value.push(e)
      })
      return _value
    },
    isQuery: function (key) { return true },
    isResponse: function (value) { return null != value  }
  })
}

/*
all these tests are sync, because it makes it really
easy to reason about the ordering of events and
produce correct tests

copied these tests from ./index.js and added explicit
ordering from the cb's but I think these cases test cases
are too simple to really have problems. This still looks
like 
*/

tape('simple', function (t) {
  var gq = GQ({query: true})
  gq.query('query', function (err, value) {
    t.equal(value, true)
    t.notOk(err)
    t.end()
  })
  check_cb(null, true)
})


tape('stream: send response', function (t) {
  var gq = GQ({})
  var stream = gq.createStream(), read_cb

  stream.sink(function (_, cb) {
    read_cb = cb
  })

  //send a request
  read_cb(null, {foo: -1})

  stream.source(null, function (err, msg) {
    if(err) throw err
    t.deepEqual(msg, {foo: ['bar']})
    t.end()
  })

  check_cb(null, ['bar'])
})

tape('stream: multiple responses', function (t) {
  t.plan(5)
  var gq = GQ({})
  var stream = gq.createStream(1), read_cb
  var stream2 = gq.createStream(2), read_cb2

  stream.sink(function (_, cb) {
    read_cb = cb
  })

  stream.source(null, function (err, msg) {
    if(err) throw err
    t.deepEqual(msg, {foo: -2})
    stream.source(null, function (err, msg) {
      if(err) throw err
      t.deepEqual(msg, {foo: ['bar']}, 'receive first update')
      console.log(gq.state, "first update, should be [bar]")
      stream.source(null, function (err, msg) {
        if(err) throw err
        t.deepEqual(msg, {foo: ['baz']}, 'recieve 2nd update')
        console.log(gq.state)
      })
    })
  })

  read_cb(null, {foo: -1})

  check_cb()

  stream2.sink(function (_, cb) {
    read_cb2 = cb
  })

  read_cb2(null, {foo: ['bar']})
  read_cb2(null, {foo: ['baz']})

  var stream3 = gq.createStream(3), read_cb3

  stream3.sink(function (_, cb) {
    read_cb3 = cb
  })

  stream3.source(null, function (err, data) {
    if(err) throw err
    t.deepEqual(data, {foo: -2})
    stream3.source(null, function (err, data) {
      t.deepEqual(data, {foo: ['bar', 'baz']}, 'receive both updates')
      t.end()
    })
  })

  read_cb3(null, {foo: -1})
})














